from __future__ import annotations
import pandas as pd
import httpx
from dataclasses import dataclass, field
import logging
from typing import List, Optional, Union
from molharbor.data import Response, ResponseSupplier
from molharbor.exceptions import LoginError
from molharbor.enums import SearchType, ResultStatus
from molharbor.utils import compound_search_payload
from pydantic import ValidationError


class Molport:
    __slots__ = ["client", "_api_key", "_username", "_password"]

    def __init__(self):
        self.client = httpx.Client()
        self._api_key = None
        self._username = None
        self._password = None

    def __repr__(self) -> str:
        return type(self).__name__ + "()"

    @property
    def credentials(self):
        if self.api_key:
            return {"api_key": self.api_key}
        elif self.username and self.password:
            return {"username": self.username, "password": self.password}
        else:
            raise LoginError(
                "No credentials are provided. Please login with username and password or api_key using .login()"
            )

    @property
    def api_key(self):
        return self._api_key

    @api_key.setter
    def api_key(self, value):
        self._api_key = value
        print("API key is set and will be used as default for all requests")
        self._username = None
        self._password = None

    @property
    def username(self):
        return self._username

    @username.setter
    def username(self, value):
        self._username = value

    @property
    def password(self):
        return self._password

    @password.setter
    def password(self, value):
        self._password = value

    def login(
        self,
        username: Optional[str] = None,
        password: Optional[str] = None,
        api_key: Optional[str] = None,
    ):
        """
        Login to Molport API. If api_key is provided, it will be used as default for all requests.
        """
        if all([username, password, api_key]):
            raise LoginError("Please provide either username and password or api_key")
        elif api_key:
            self._api_key = api_key
            self._username = None
            self._password = None
        elif username and password:
            self._username = username
            self._password = password
            self._api_key = None
        else:
            raise LoginError(
                "Please provide either username and password or api_key to login"
            )

    def find(
        self,
        /,
        smiles: str,
        *,
        search_type: Union[SearchType, int] = SearchType.EXACT_FRAGMENT,
        max_search_time: Optional[int] = None,
        max_results: int = 10000,
        similarity: float = 0.9,
        return_response: bool = False,
    ) -> List[MolportCompound] | Response:
        """Find compounds by SMILES string in Molport database, have the same default values as the API

        Args:
            smiles (str): SMILES string of the compound
            search_type (Union[SearchType, int], optional): _description_. Defaults to SearchType.EXACT_FRAGMENT.
            max_search_time (Optional[int], optional): time in miliseconds - maximum search time to be spent on chemical search
            max_results (int, optional): maximum result count which must be returned as result; currently maximum allowed value is 10000. Defaults to 10000.
            similarity (float, optional): if similarity search is made, it is possible to provide similarity index in range 0 - 1. Defaults to 0.9.
            return_response (bool, optional): If True, returns the response object. Otherwise parses the response and returns a list of `MolportCompound` objects. Defaults to False.

        Raises:
            TypeError: If SMILES is not a string
            LoginError: If credentials are incorrect
            ValidationError: If the response or payload is not valid

        Returns:
            List[MolportCompound] | Response: List of MolportCompound objects or Response object
        """
        if not isinstance(smiles, str):
            raise TypeError("SMILES must be a string")
        payload = compound_search_payload(
            smiles=smiles,
            search_type=search_type,
            maximum_search_time=max_search_time,
            max_results=max_results,
            similarity=similarity,
            credentials=self.credentials,
        )
        similarity_request = self.client.post(
            "https://api.molport.com/api/chemical-search/search", json=payload
        )
        if similarity_request.status_code != 200:
            logging.error(f"Error code: {similarity_request.status_code}")
            return []
        try:
            response = Response(**similarity_request.json())
            if response.result.status != ResultStatus.SUCCESS.value:
                msg = response.result.message
                if "Username or password is incorrect!" in msg:
                    raise LoginError("Credentials are incorrect, please login again")
                logging.error(msg)
                return []
        except ValidationError as e:
            logging.error(e)
            return []
        if return_response:
            return response
        mols = response.data.molecules
        if not mols:
            return []
        return [MolportCompound(mol.smiles, mol.molport_id) for mol in mols]

    def get_suppliers(
        self, molport_id: str, return_response: bool = False
    ) -> Union[pd.DataFrame, ResponseSupplier]:
        """Get suppliers for a given Molport ID

        Args:
            molport_id (str): Molport ID of the compound
            return_response (bool, optional): If True, returns the response object. Otherwise parses the response and returns a DataFrame. Defaults to False.

        Raises:
            ValueError: If the response status is not 200

        Returns:
            Union[pd.DataFrame, ResponseSupplier]: DataFrame with supplier information or Response object
        """
        credentials = self.credentials
        url = "https://api.molport.com/api/molecule/load?molecule={}"
        if "api_key" in credentials:
            url += "&apikey={}"
            url = url.format(molport_id, credentials["api_key"])
        else:
            url += "&username={}&authenticationcode={}"
            url = url.format(molport_id, self.username, self.password)
        response = self.client.get(url)
        if response.status_code != 200:
            raise ValueError(f"Error code: {response.status_code}\n{response.text}")
        data = ResponseSupplier(**response.json())
        if return_response:
            return data
        else:
            return self.extract_suppliers(data)

    def extract_suppliers(self, response: ResponseSupplier) -> pd.DataFrame:
        """Extract suppliers from the response data

        Args:
            response (ResponseSupplier): Response data from the API

        Raises:
            ValueError: If the response status is not SUCCESS

        Returns:
            pd.DataFrame: DataFrame with supplier information
        """
        if response.result.status != ResultStatus.SUCCESS.value:
            raise ValueError(response.result.message)
        types = [
            "screening_block_suppliers",
            "building_block_suppliers",
            "virtual_suppliers",
        ]
        records = []
        for supp_type in types:
            if hasattr(response.data.molecule.catalogues, supp_type):
                for supp in response.data.molecule.catalogues.screening_block_suppliers:
                    for catalog in supp.catalogues:
                        data = catalog.model_dump()
                        packings = data.pop("available_packings")
                        name = {
                            "supplier_name": supp.supplier_name,
                            "supplier_type": supp_type,
                        }
                        for packing in packings:
                            record = {**name, **packing, **data}
                            records.append(record)
        df = pd.DataFrame(records)
        return df


@dataclass
class MolportCompound:
    smiles: str
    molport_id: str
    link: str = field(init=False)

    def __post_init__(self):
        self.link = (
            f"https://www.molport.com/shop/compound/{self.molport_id}"
            if self.molport_id
            else ""
        )



from molharbor.exceptions import UnknownSearchTypeException
from molharbor.data import SearchPayload
from molharbor.enums import SearchType
from typing import Dict, Optional, Union


def compound_search_payload(
    smiles: str,
    search_type: Union[SearchType, int] = SearchType.EXACT_FRAGMENT,
    maximum_search_time: Optional[int] = None,
    max_results: int = 10000,
    similarity: float = 0.9,
    credentials: Dict[str, str] = {},
) -> Dict[str, Union[str, int, float]]:
    try:
        search_type = SearchType(search_type)
    except ValueError:
        raise UnknownSearchTypeException(search_type)

    search_payload = SearchPayload(
        smiles=smiles,
        search_type=search_type,
        maximum_search_time=maximum_search_time,
        max_results=max_results,
        similarity=similarity,
        **credentials,
    )
    return search_payload.model_dump(by_alias=True, exclude_none=True)



from molharbor.enums import SearchType
from typing import Union


class UnknownSearchTypeException(ValueError):
    """Exception raised when an unknown search type is specified."""

    def __init__(self, search_type: Union[SearchType, int]) -> None:
        self.search_type = search_type
        super().__init__(
            f"Unknown search type: {search_type}. Expected one of: {list(SearchType)}"
        )


class LoginError(Exception):
    """Exception raised when login fails."""

    pass



from enum import Enum


class ResultStatus(Enum):
    SUCCESS = 1
    ERROR = 2


class SearchType(Enum):
    SUBSTRUCTURE = 1
    SUPERSTRUCTURE = 2
    EXACT = 3
    SIMILARITY = 4
    PERFECT = 5
    EXACT_FRAGMENT = 6



from typing import List, Optional
from pydantic_core import PydanticCustomError
from pydantic import BaseModel, ConfigDict, Field, model_validator
import numpy as np
from molharbor.enums import SearchType


class Molecule(BaseModel):
    id: int = Field(..., alias="Id")
    molport_id: str = Field(..., alias="MolPort Id")
    smiles: str = Field(None, alias="SMILES")
    canonical_smiles: str = Field(None, alias="Canonical SMILES")
    verified_amount: float = Field(None, alias="Verified Amount")
    unverified_amount: float = Field(None, alias="Unverified Amount")
    similarity_index: float = Field(None, alias="Similarity Index")


class Data(BaseModel):
    molecules: Optional[List[Molecule]] = Field(alias="Molecules")
    version: str = Field(alias="Version")


class Result(BaseModel):
    status: int = Field(alias="Status")
    message: str = Field(alias="Message")


class Response(BaseModel):
    result: Result = Field(alias="Result")
    data: Data = Field(alias="Data")


class AvailablePacking(BaseModel):
    amount: float = Field(np.nan, alias="Amount")
    measure: str = Field(alias="Measure")
    measure_id: int = Field(alias="Measure Id")
    price: float = Field(alias="Price")
    currency: str = Field(alias="Currency")
    currency_id: int = Field(alias="Currency Id")
    delivery_days: int = Field(alias="Delivery Days")


class Catalog(BaseModel):
    catalog_id: int = Field(alias="Catalog Id")
    catalog_number: str = Field(alias="Catalog Number")
    stock: Optional[float] = Field(None, alias="Stock")
    stock_measure: Optional[str] = Field(None, alias="Stock Measure")
    stock_measure_id: Optional[int] = Field(None, alias="Stock Measure Id")
    purity: str = Field("unknown", alias="Purity")
    last_update_date: str = Field(alias="Last Update Date")
    last_update_date_exact: str = Field(alias="Last Update Date Exact")
    available_packings: List[AvailablePacking] = Field(alias="Available Packings")


class Supplier(BaseModel):
    supplier_name: str = Field(alias="Supplier Name")
    supplier_id: int = Field(alias="Supplier Id")
    minimum_order: int = Field(alias="Minimum Order")
    currency: str = Field(alias="Currency")
    currency_id: int = Field(alias="Currency Id")
    catalogues: List[Catalog] = Field(alias="Catalogues")


class BagOfSuppliers(BaseModel):
    screening_block_suppliers: Optional[List[Supplier]] = Field(
        alias="Screening Block Suppliers"
    )
    building_block_suppliers: Optional[List[Supplier]] = Field(
        alias="Building Block Suppliers"
    )
    virtual_suppliers: Optional[List[Supplier]] = Field(alias="Virtual Suppliers")


class ShipmentCost(BaseModel):
    price: float = Field(alias="Price")
    currency: str = Field(alias="Currency")
    currency_Id: int = Field(alias="Currency Id")
    location_Type: str = Field(alias="Location Type")
    location: str = Field(alias="Location")
    location_id: int = Field(alias="Location Id")
    shipment_type: str = Field(alias="Shipment Type")
    shipment_type_id: int = Field(alias="Shipment Type Id")


class Molecule2(BaseModel):
    id: int = Field(..., alias="Id")
    molport_id: str = Field(..., alias="Molport Id")
    smiles: str = Field(..., alias="SMILES")
    canonical_smiles: str = Field(..., alias="Canonical SMILES")
    iupac: str = Field(..., alias="IUPAC")
    formula: str = Field(..., alias="Formula")
    molecular_weight: float = Field(..., alias="Molecular Weight")
    status: str = Field(..., alias="Status")
    type: str = Field(..., alias="Type")
    targest_stock: str = Field(..., alias="Largest Stock")
    largest_stock_measure: str = Field(..., alias="Largest Stock Measure")
    largest_stock_measure_id: int = Field(..., alias="Largest Stock Measure Id")
    catalogues: BagOfSuppliers = Field(..., alias="Catalogues")
    synonyms: List[str] = Field(..., alias="Synonyms")


class DataSupplier(BaseModel):
    molecule: Molecule2 = Field(None, alias="Molecule")
    version: str = Field(alias="Version")


class ResponseSupplier(Response):
    data: DataSupplier = Field(alias="Data")


class SearchPayload(BaseModel):
    smiles: str = Field(..., serialization_alias="Structure")
    search_type: SearchType = Field(
        SearchType.EXACT_FRAGMENT, serialization_alias="Search Type"
    )
    maximum_search_time: Optional[int] = Field(
        None, serialization_alias="Maximum Search Time"
    )
    max_results: int = Field(10000, serialization_alias="Maximum Result Count")
    similarity: float = Field(0.9, serialization_alias="Chemical Similarity Index")
    api_key: Optional[str] = Field(None, serialization_alias="API Key")
    username: Optional[str] = Field(None, serialization_alias="User Name")
    password: Optional[str] = Field(None, serialization_alias="Authentication Code")
    model_config = ConfigDict(use_enum_values=True)

    @model_validator(mode="before")
    def check_auth(cls, values: dict):
        api_key, username, password = (
            values.get("api_key"),
            values.get("username"),
            values.get("password"),
        )
        if api_key:
            values.pop("username", None)
            values.pop("password", None)
            return values
        elif not all((username, password)):
            raise PydanticCustomError(
                "Either username and password or api_key must be provided",
                f"Wrong credintials combination: api_key={api_key}, username={username}, password={password}",
            )
        values.pop("api_key", None)
        return values
